# Generated by Django 2.0rc1 on 2017-11-19 08:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_appointment", "0008_auto_20171115_1601")]

    operations = [
        migrations.AlterField(
            model_name="appointment",
            name="appt_reason",
            field=models.CharField(
                choices=[
                    ("scheduled", "Routine / Scheduled"),
                    ("unscheduled", "Unscheduled"),
                ],
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_reason",
            field=models.CharField(
                choices=[
                    ("scheduled", "Routine / Scheduled"),
                    ("unscheduled", "Unscheduled"),
                ],
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
    ]



# Generated by Django 3.0.9 on 2020-09-11 01:25

import edc_sites.models
from django.db import migrations

import edc_appointment.managers


class Migration(migrations.Migration):

    dependencies = [
        ("edc_appointment", "0023_auto_20200821_2119"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="appointment",
            managers=[
                ("on_site", edc_sites.models.CurrentSiteManager()),
                ("objects", edc_appointment.managers.AppointmentManager()),
            ],
        ),
    ]



from edc_appointment.constants import (
    COMPLETE_APPT,
    IN_PROGRESS_APPT,
    INCOMPLETE_APPT,
    MISSED_APPT,
    NEW_APPT,
)


def constants(request) -> dict:
    dct = dict(
        COMPLETE_APPT=COMPLETE_APPT,
        INCOMPLETE_APPT=INCOMPLETE_APPT,
        IN_PROGRESS_APPT=IN_PROGRESS_APPT,
        MISSED_APPT=MISSED_APPT,
        NEW_APPT=NEW_APPT,
    )
    return dct



from django.db.models import ProtectedError
from django.db.models.signals import post_save
from django.test import TestCase
from edc_facility.import_holidays import import_holidays
from edc_protocol import Protocol
from edc_reference import site_reference_configs
from edc_visit_schedule import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED

from edc_appointment_app.models import SubjectVisit
from edc_appointment_app.visit_schedule import visit_schedule1, visit_schedule2

from ...constants import INCOMPLETE_APPT
from ...creators import UnscheduledAppointmentCreator
from ...managers import AppointmentDeleteError
from ...models import Appointment, appointments_on_post_delete
from ...utils import delete_appointment_in_sequence
from ..helper import Helper


class TestDeleteAppointment(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        post_save.connect(
            appointments_on_post_delete, dispatch_uid="appointments_on_post_delete"
        )

    def setUp(self):
        post_save.disconnect(dispatch_uid="appointments_on_post_delete")
        self.subject_identifier = "12345"
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_visit_schedules.register(visit_schedule=visit_schedule2)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=Protocol().study_open_datetime,
        )
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_appointment_app.subjectvisit"}
        )
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.filter(subject_identifier=self.subject_identifier)
        self.assertEqual(appointments.count(), 4)

        appointment = Appointment.objects.get(timepoint=0.0)
        SubjectVisit.objects.create(
            appointment=appointment,
            subject_identifier=appointment.subject_identifier,
            report_datetime=appointment.appt_datetime,
            visit_schedule_name=visit_schedule1.name,
            schedule_name="schedule1",
            visit_code="1000",
            reason=SCHEDULED,
        )
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save_base(update_fields=["appt_status"])
        appointment.refresh_from_db()

        for i in range(1, 4):
            creator = UnscheduledAppointmentCreator(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=visit_schedule1.name,
                schedule_name="schedule1",
                visit_code="1000",
                # appt_datetime=appointment.appt_datetime + relativedelta(days=i),
                visit_code_sequence=appointment.visit_code_sequence + i,
                timepoint=appointment.timepoint,
            )
            creator.appointment.appt_status = INCOMPLETE_APPT
            creator.appointment.save_base(update_fields=["appt_status"])

        self.appt_datetimes = [
            o.appt_datetime for o in Appointment.objects.all().order_by("appt_datetime")
        ]

    def test_delete_0_appointment_in_sequence(self):
        self.assertEqual(
            [0, 1, 2, 3],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=0)
        # raises ProtectedError because Subjectvisit exists
        self.assertRaises(ProtectedError, delete_appointment_in_sequence, appointment)
        SubjectVisit.objects.get(appointment=appointment).delete()
        # raises AppointmentDeleteError (from manager) because not allowed by manager
        self.assertRaises(AppointmentDeleteError, delete_appointment_in_sequence, appointment)
        # assert nothing was done
        self.assertEqual(
            [0, 1, 2, 3],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )

    def test_delete_first_appointment_in_sequence(self):
        self.assertEqual(
            [0, 1, 2, 3],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=1)
        delete_appointment_in_sequence(appointment)
        self.assertEqual(
            [0, 1, 2],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )

    def test_delete_second_appointment_in_sequence(self):
        self.assertEqual(
            [0, 1, 2, 3],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=2)
        delete_appointment_in_sequence(appointment)
        self.assertEqual(
            [0, 1, 2],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )

    def test_delete_third_appointment_in_sequence(self):
        self.assertEqual(
            [0, 1, 2, 3],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=3)
        delete_appointment_in_sequence(appointment)
        self.assertEqual(
            [0, 1, 2],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )



# Generated by Django 2.0b1 on 2017-11-15 16:01

import _socket
import edc_model_fields.fields.userfield
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_appointment", "0007_auto_20170302_2220")]

    operations = [
        migrations.AlterModelOptions(
            name="historicalappointment",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical ",
            },
        ),
        migrations.RemoveField(model_name="historicalappointment", name="visit_instance"),
        migrations.AddField(
            model_name="appointment",
            name="device_created",
            field=models.CharField(blank=True, max_length=10),
        ),
        migrations.AddField(
            model_name="appointment",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10),
        ),
        migrations.AddField(
            model_name="historicalappointment",
            name="device_created",
            field=models.CharField(blank=True, max_length=10),
        ),
        migrations.AddField(
            model_name="historicalappointment",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10),
        ),
        migrations.AddField(
            model_name="historicalappointment",
            name="history_change_reason",
            field=models.CharField(max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="appt_reason",
            field=models.CharField(
                blank=True,
                choices=[
                    ("scheduled", "Routine / Scheduled"),
                    ("unscheduled", "Unscheduled"),
                ],
                help_text="Reason for appointment",
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="appt_type",
            field=models.CharField(
                choices=[
                    ("clinic", "In clinic"),
                    ("home", "At home"),
                    ("hospital", "In hospital"),
                    ("telephone", "By telephone"),
                ],
                default="clinic",
                help_text="Default for subject may be edited Subject Configuration.",
                max_length=20,
                verbose_name="Appointment type",
            ),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
            ),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="user_created",
            field=edc_model_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="user_modified",
            field=edc_model_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_reason",
            field=models.CharField(
                blank=True,
                choices=[
                    ("scheduled", "Routine / Scheduled"),
                    ("unscheduled", "Unscheduled"),
                ],
                help_text="Reason for appointment",
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_type",
            field=models.CharField(
                choices=[
                    ("clinic", "In clinic"),
                    ("home", "At home"),
                    ("hospital", "In hospital"),
                    ("telephone", "By telephone"),
                ],
                default="clinic",
                help_text="Default for subject may be edited Subject Configuration.",
                max_length=20,
                verbose_name="Appointment type",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="user_created",
            field=edc_model_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="user_modified",
            field=edc_model_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.RemoveField(model_name="appointment", name="visit_instance"),
        migrations.AlterUniqueTogether(
            name="appointment",
            unique_together={
                (
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "timepoint",
                    "visit_code_sequence",
                )
            },
        ),
    ]



from typing import Optional

from dateutil.relativedelta import MO, relativedelta
from django.contrib import admin
from django.db.models import QuerySet
from django.utils.translation import gettext as _
from edc_utils import get_utcnow

from ..constants import (
    FUTURE_DATE,
    LAST_WEEK,
    NEXT_WEEK,
    NOT_NULL,
    PAST_DATE,
    THIS_WEEK,
    TODAY,
    TOMORROW,
)


class AppointmentListFilter(admin.SimpleListFilter):
    title = _("Appointment date")

    parameter_name = "appt_datetime"
    field_name = "appt_datetime"

    def lookups(self, request, model_admin):
        return (
            (TODAY, "Today"),
            (TOMORROW, "Tomorrow"),
            (THIS_WEEK, "This week"),
            (NEXT_WEEK, "Next week"),
            (FUTURE_DATE, "Any future date"),
            (PAST_DATE, "Any past date"),
            (NOT_NULL, "Has date"),
        )

    @property
    def extra_queryset_options(self) -> dict:
        return {}

    def queryset(self, request, queryset) -> Optional[QuerySet]:
        morning = get_utcnow().replace(second=0, hour=0, minute=0)
        monday = morning + relativedelta(weekday=MO(-1))
        night = get_utcnow().replace(second=59, hour=23, minute=59)
        qs = None
        if self.value() == NEXT_WEEK:
            qs = queryset.filter(
                **{
                    f"{self.field_name}__gte": monday + relativedelta(weeks=1),
                    f"{self.field_name}__lt": monday + relativedelta(weeks=2),
                },
                **self.extra_queryset_options,
            ).order_by(self.field_name)
        if self.value() == THIS_WEEK:
            qs = queryset.filter(
                **{
                    f"{self.field_name}__gte": monday,
                    f"{self.field_name}__lt": monday + relativedelta(weeks=1),
                },
                **self.extra_queryset_options,
            ).order_by(self.field_name)
        if self.value() == TODAY:
            qs = queryset.filter(
                **{
                    f"{self.field_name}__gte": morning,
                    f"{self.field_name}__lt": night,
                },
                **self.extra_queryset_options,
            ).order_by(self.field_name)
        if self.value() == TOMORROW:
            qs = queryset.filter(
                **{
                    f"{self.field_name}__gte": night,
                    f"{self.field_name}__lt": night + relativedelta(days=1),
                },
                **self.extra_queryset_options,
            ).order_by(self.field_name)
        if self.value() == LAST_WEEK:
            qs = queryset.filter(
                **{
                    f"{self.field_name}__gte": monday,
                    f"{self.field_name}__lt": monday - relativedelta(weeks=1),
                },
                **self.extra_queryset_options,
            ).order_by(f"-{self.field_name}")
        if self.value() == PAST_DATE:
            qs = queryset.filter(
                **{f"{self.field_name}__lt": morning},
                **self.extra_queryset_options,
            ).order_by(f"-{self.field_name}")
        if self.value() == FUTURE_DATE:
            qs = queryset.filter(
                **{f"{self.field_name}__gt": night},
                **self.extra_queryset_options,
            ).order_by(self.field_name)
        if self.value() == NOT_NULL:
            qs = queryset.filter(
                **{f"{self.field_name}__isnull": False},
                **self.extra_queryset_options,
            ).order_by(self.field_name)
        return qs



from django.core.exceptions import ObjectDoesNotExist
from django.db.models.signals import post_delete, post_save, pre_delete
from django.dispatch import receiver
from edc_utils import formatted_datetime, get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.model_mixins import VisitModelMixin

from ..appointment_status_updater import (
    AppointmentStatusUpdater,
    AppointmentStatusUpdaterError,
)
from ..constants import IN_PROGRESS_APPT
from ..managers import AppointmentDeleteError
from ..model_mixins import AppointmentModelMixin
from ..models import Appointment
from ..utils import (
    cancelled_appointment,
    get_appointment_model_name,
    missed_appointment,
    update_unscheduled_appointment_sequence,
)


@receiver(post_save, sender=Appointment, weak=False, dispatch_uid="appointment_post_save")
def appointment_post_save(sender, instance, raw, update_fields, **kwargs):
    if not raw:
        if not update_fields:
            # use the AppointmentStatusUpdater to set all
            # other appointments to be NOT in progress
            try:
                if instance.appt_status == IN_PROGRESS_APPT:
                    AppointmentStatusUpdater(instance, clear_others_in_progress=True)
            except AttributeError as e:
                if "appt_status" not in str(e):
                    raise
            except AppointmentStatusUpdaterError:
                pass
        # try to create_missed_visit_from_appointment
        # if appt_status == missed
        missed_appointment(instance)

        # try to delete subject visit
        # if appt status = CANCELLED_APPT
        cancelled_appointment(instance)


@receiver(post_save, weak=False, dispatch_uid="create_appointments_on_post_save")
def create_appointments_on_post_save(sender, instance, raw, created, using, **kwargs):
    """Method `Model.create_appointments` is not typically used.

    See schedule.put_on_schedule() in edc_visit_schedule.
    """
    if not raw and not kwargs.get("update_fields"):
        try:
            instance.create_appointments()
        except AttributeError as e:
            if "create_appointments" not in str(e):
                raise


@receiver(post_save, weak=False, dispatch_uid="update_appt_status_on_subject_visit_post_save")
def update_appt_status_on_subject_visit_post_save(
    sender, instance, raw, update_fields, **kwargs
):
    if not raw and not update_fields:
        if isinstance(instance, (AppointmentModelMixin, VisitModelMixin)):
            try:
                AppointmentStatusUpdater(
                    instance.appointment,
                    change_to_in_progress=True,
                    clear_others_in_progress=True,
                )
            except AttributeError:
                pass
            except AppointmentStatusUpdaterError:
                pass


@receiver(
    pre_delete, sender=Appointment, weak=False, dispatch_uid="appointments_on_pre_delete"
)
def appointments_on_pre_delete(sender, instance, using, **kwargs):
    if instance.visit_code_sequence == 0:
        schedule = site_visit_schedules.get_visit_schedule(
            instance.visit_schedule_name
        ).schedules.get(instance.schedule_name)
        onschedule_datetime = schedule.onschedule_model_cls.objects.get(
            subject_identifier=instance.subject_identifier
        ).onschedule_datetime
        try:
            offschedule_datetime = schedule.offschedule_model_cls.objects.get(
                subject_identifier=instance.subject_identifier
            ).offschedule_datetime
        except ObjectDoesNotExist:
            raise AppointmentDeleteError(
                f"Appointment may not be deleted. "
                f"Subject {instance.subject_identifier} is on schedule "
                f"'{instance.visit_schedule.verbose_name}.{instance.schedule_name}' "
                f"as of '{formatted_datetime(onschedule_datetime)}'. "
                f"Got appointment {instance.visit_code}.{instance.visit_code_sequence} "
                f"datetime {formatted_datetime(instance.appt_datetime)}. "
                f"Perhaps complete off schedule model "
                f"'{instance.schedule.offschedule_model_cls().verbose_name.title()}' "
                f"first."
            )
        else:
            if onschedule_datetime <= instance.appt_datetime <= offschedule_datetime:
                raise AppointmentDeleteError(
                    f"Appointment may not be deleted. "
                    f"Subject {instance.subject_identifier} is on schedule "
                    f"'{instance.visit_schedule.verbose_name}.{instance.schedule_name}' "
                    f"as of '{formatted_datetime(onschedule_datetime)}' "
                    f"until '{formatted_datetime(get_utcnow())}'. "
                    f"Got appointment datetime "
                    f"{formatted_datetime(instance.appt_datetime)}. "
                )
    else:
        # TODO: any conditions for unscheduled?
        pass


@receiver(
    post_delete, sender=Appointment, weak=False, dispatch_uid="appointments_on_post_delete"
)
def appointments_on_post_delete(sender, instance, using, **kwargs):
    # Holiday model comes in a sender, why ???
    if (
        not kwargs.get("update_fields")
        and sender._meta.label_lower == get_appointment_model_name()
    ):
        update_unscheduled_appointment_sequence(subject_identifier=instance.subject_identifier)



from datetime import date

from django.db import models
from django.db.models.deletion import PROTECT
from edc_consent.model_mixins import RequiresConsentFieldsModelMixin
from edc_crf.model_mixins import CrfModelMixin
from edc_identifier.managers import SubjectIdentifierManager
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_list_data.model_mixins import ListModelMixin
from edc_metadata.model_mixins.creates import CreatesMetadataModelMixin
from edc_model.models import BaseUuidModel
from edc_offstudy.model_mixins import OffstudyModelMixin
from edc_reference.model_mixins import ReferenceModelMixin
from edc_registration.model_mixins import UpdatesOrCreatesRegistrationModelMixin
from edc_screening.model_mixins import ScreeningModelMixin
from edc_sites.models import SiteModelMixin
from edc_utils import get_utcnow
from edc_visit_schedule.model_mixins import (
    OffScheduleModelMixin,
    OnScheduleModelMixin,
    VisitCodeFieldsModelMixin,
)
from edc_visit_tracking.model_mixins import (
    SubjectVisitMissedModelMixin,
    VisitModelMixin,
)
from edc_visit_tracking.models import SubjectVisitMissedReasons

from edc_appointment.models import Appointment


class Panel(ListModelMixin):
    class Meta:
        pass


class SubjectVisit(
    SiteModelMixin,
    VisitModelMixin,
    ReferenceModelMixin,
    CreatesMetadataModelMixin,
    RequiresConsentFieldsModelMixin,
    BaseUuidModel,
):
    appointment = models.OneToOneField(
        Appointment,
        on_delete=PROTECT,  # related_name="test_visit_schedule_appointment"
    )

    subject_identifier = models.CharField(max_length=25, null=True)

    report_datetime = models.DateTimeField()

    reason = models.CharField(max_length=25, null=True)


class SubjectRequisition(
    SiteModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    VisitCodeFieldsModelMixin,
    BaseUuidModel,
):
    @classmethod
    def related_visit_model_attr(cls):
        return "subject_visit"

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=models.PROTECT, related_name="+")

    panel = models.ForeignKey(Panel, on_delete=PROTECT)

    class Meta(BaseUuidModel.Meta):
        pass


class SubjectVisitMissed(SiteModelMixin, SubjectVisitMissedModelMixin, BaseUuidModel):

    subject_visit = models.OneToOneField(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    missed_reasons = models.ManyToManyField(
        SubjectVisitMissedReasons, blank=True, related_name="missed_reasons"
    )

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Missed Visit Report"
        verbose_name_plural = "Missed Visit Report"


class SubjectConsent(
    SiteModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    BaseUuidModel,
):
    report_datetime = models.DateTimeField(default=get_utcnow)

    consent_datetime = models.DateTimeField(default=get_utcnow)

    version = models.CharField(max_length=25, default="1")

    identity = models.CharField(max_length=25)

    confirm_identity = models.CharField(max_length=25)

    dob = models.DateField(default=date(1995, 1, 1))


class SubjectScreening(ScreeningModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class SubjectOffstudy(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class SubjectOffstudy2(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class SubjectOffstudyFive(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class SubjectOffstudySix(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class SubjectOffstudySeven(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class DeathReport(SiteModelMixin, BaseUuidModel):
    subject_identifier = models.CharField(max_length=25, null=True)

    report_datetime = models.DateTimeField()


# visit_schedule


class OnSchedule(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffSchedule(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    pass


class OnScheduleOne(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleOne(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    class Meta(OffScheduleModelMixin.Meta):
        pass


class OnScheduleTwo(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleTwo(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    pass


class OnScheduleThree(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleThree(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    pass


class CrfOne(CrfModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    f1 = models.CharField(max_length=50, null=True, blank=True)

    f2 = models.CharField(max_length=50, null=True, blank=True)

    f3 = models.CharField(max_length=50, null=True, blank=True)


class CrfTwo(CrfModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    f1 = models.CharField(max_length=50, null=True, blank=True)

    f2 = models.CharField(max_length=50, null=True, blank=True)

    f3 = models.CharField(max_length=50, null=True, blank=True)


class CrfThree(CrfModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    f1 = models.CharField(max_length=50, null=True, blank=True)

    f2 = models.CharField(max_length=50, null=True, blank=True)

    f3 = models.CharField(max_length=50, null=True, blank=True)


class CrfFour(CrfModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    f1 = models.CharField(max_length=50, null=True, blank=True)

    f2 = models.CharField(max_length=50, null=True, blank=True)

    f3 = models.CharField(max_length=50, null=True, blank=True)


class CrfFive(CrfModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    f1 = models.CharField(max_length=50, null=True, blank=True)

    f2 = models.CharField(max_length=50, null=True, blank=True)

    f3 = models.CharField(max_length=50, null=True, blank=True)



from datetime import datetime
from decimal import Decimal
from zoneinfo import ZoneInfo

from dateutil.relativedelta import relativedelta
from django.core.exceptions import ObjectDoesNotExist
from django.forms import ValidationError
from django.test import TestCase, override_settings
from edc_facility.import_holidays import import_holidays
from edc_form_validators import ModelFormFieldValidatorError
from edc_metadata import KEYED, REQUIRED
from edc_metadata.models import CrfMetadata, RequisitionMetadata
from edc_reference import site_reference_configs
from edc_visit_schedule import site_visit_schedules
from edc_visit_schedule.constants import DAY01
from edc_visit_tracking.constants import SCHEDULED
from edc_visit_tracking.model_mixins import PreviousVisitError
from edc_visit_tracking.utils import get_subject_visit_missed_model_cls

from edc_appointment_app.models import SubjectVisit
from edc_appointment_app.visit_schedule import visit_schedule1, visit_schedule2

from ...constants import (
    IN_PROGRESS_APPT,
    MISSED_APPT,
    ONTIME_APPT,
    SCHEDULED_APPT,
    UNSCHEDULED_APPT,
)
from ...form_validators import (
    INVALID_APPT_TIMING_CRFS_EXIST,
    INVALID_APPT_TIMING_REQUISITIONS_EXIST,
    INVALID_PREVIOUS_VISIT_MISSING,
    AppointmentFormValidator,
)
from ...form_validators.appointment_form_validator import (
    INVALID_MISSED_APPT_NOT_ALLOWED,
    INVALID_MISSED_APPT_NOT_ALLOWED_AT_BASELINE,
)
from ...models import Appointment
from ...utils import get_previous_appointment
from ..appointment_test_case_mixin import AppointmentTestCaseMixin
from ..helper import Helper


class TestAppointmentFormValidator(AppointmentTestCaseMixin, TestCase):

    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        self.subject_identifier = "12345"
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_visit_schedules.register(visit_schedule=visit_schedule2)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=datetime(2017, 1, 7, tzinfo=ZoneInfo("UTC")),
        )
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_appointment_app.subjectvisit"}
        )

    def test_get_previous(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all()
        for i in [0, 1]:
            Appointment.objects.create(
                subject_identifier=appointments[i].subject_identifier,
                appt_datetime=appointments[i].appt_datetime + relativedelta(hours=i + 1),
                timepoint=appointments[i].timepoint,
                visit_code=appointments[i].visit_code,
                visit_code_sequence=i + 1,
                visit_schedule_name=appointments[i].visit_schedule_name,
                schedule_name=appointments[i].schedule_name,
            )
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(
            [f"{obj.visit_code}.{obj.visit_code_sequence}" for obj in appointments],
            ["1000.0", "1000.1", "1000.2", "2000.0", "3000.0", "4000.0"],
        )

        self.assertIsNone(get_previous_appointment(appointments[0]))

        self.assertEqual(
            appointments[4], get_previous_appointment(appointments[5], include_interim=True)
        )
        self.assertEqual(appointments[4], get_previous_appointment(appointments[5]))

        self.assertEqual(
            appointments[2], get_previous_appointment(appointments[3], include_interim=True)
        )
        self.assertEqual(appointments[0], get_previous_appointment(appointments[3]))

        self.assertEqual(
            appointments[3], get_previous_appointment(appointments[4], include_interim=True)
        )
        self.assertEqual(appointments[3], get_previous_appointment(appointments[4]))

    def test_(self):
        try:
            AppointmentFormValidator(cleaned_data={})
        except ModelFormFieldValidatorError as e:
            self.fail(f"ModelFormFieldValidatorError unexpectedly raised. Got {e}")

    def test_visit_report_sequence(self):
        """Asserts a sequence error is raised if previous visit
        not complete for an in progress appointment.
        """
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all()

        # try to add second appt before the first
        # should fail
        form_validator = AppointmentFormValidator(
            cleaned_data=dict(appt_status=IN_PROGRESS_APPT), instance=appointments[1]
        )
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate_visit_report_sequence()
        self.assertIn(INVALID_PREVIOUS_VISIT_MISSING, form_validator._error_codes)
        self.assertIn("1000.0", str(cm.exception))

        # try to add second appt where first visit report is complete
        # should succeed
        # add a visit 0
        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        form_validator = AppointmentFormValidator(
            cleaned_data=dict(appt_status=IN_PROGRESS_APPT), instance=appointments[1]
        )
        try:
            form_validator.validate_visit_report_sequence()
        except ValidationError:
            self.fail("ValidationError unexpectedly raised.")

        # try to manually add a visit 2 before visit 1
        # should fail
        self.assertRaises(
            PreviousVisitError,
            SubjectVisit.objects.create,
            appointment=appointments[2],
            report_datetime=appointments[2].appt_datetime,
            reason=SCHEDULED,
        )

    def test_visit_report_sequence2(self):
        """Asserts a sequence error is raised if previous visit
        not complete for an in progress appointment.

        Validate the visit_code_sequence
        """
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")

        # add continuation appt (visit_code_sequence=1)
        for i in [0, 1]:
            Appointment.objects.create(
                subject_identifier=appointments[i].subject_identifier,
                appt_datetime=appointments[i].appt_datetime + relativedelta(hours=i + 1),
                timepoint=appointments[i].timepoint,
                visit_code=appointments[i].visit_code,
                visit_code_sequence=i + 1,
                visit_schedule_name=appointments[i].visit_schedule_name,
                schedule_name=appointments[i].schedule_name,
                appt_reason=UNSCHEDULED_APPT,
            )

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")

        # appointments are
        # 1000.0, 1000.1, 1000.2 2000.0, 3000.0, 4000.0
        # NO visit reports.
        self.assertEqual(appointments[1].visit_code, "1000")
        self.assertEqual(appointments[1].visit_code_sequence, 1)

        # try to add second appt before the first
        # should fail
        form_validator = AppointmentFormValidator(
            cleaned_data=dict(appt_status=IN_PROGRESS_APPT), instance=appointments[1]
        )
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate_visit_report_sequence()
        self.assertIn(INVALID_PREVIOUS_VISIT_MISSING, form_validator._error_codes)
        self.assertIn("1000.0", str(cm.exception))

        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )

        # 1000.0 (reported), 1000.1, 1000.2 2000.0, 3000.0, 4000.0
        # try to add 1000.2
        # should fail because 1000.1 not reported

        form_validator = AppointmentFormValidator(
            cleaned_data=dict(appt_status=IN_PROGRESS_APPT), instance=appointments[2]
        )
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate_visit_report_sequence()
        self.assertIn("1000.1", str(cm.exception))
        self.assertIn(INVALID_PREVIOUS_VISIT_MISSING, form_validator._error_codes)

        SubjectVisit.objects.create(
            appointment=appointments[1],
            report_datetime=appointments[1].appt_datetime,
            reason=SCHEDULED,
        )

        form_validator = AppointmentFormValidator(
            cleaned_data=dict(appt_status=IN_PROGRESS_APPT), instance=appointments[2]
        )
        try:
            form_validator.validate_visit_report_sequence()
        except ValidationError:
            self.fail("ValidationError unexpectedly raised.")

    def test_interim_sequence(self):
        self.helper.consent_and_put_on_schedule()

        form_validator = AppointmentFormValidator(
            cleaned_data=dict(
                subject_identifier=self.subject_identifier,
                appt_status=IN_PROGRESS_APPT,
                visit_code="1000",
                visit_code_sequence=1,
                timepoint=Decimal("0.1"),
            )
        )
        form_validator.validate_visit_report_sequence()

    def test_confirm_appt_field_attrs(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].visit_code, DAY01)
        self.assertEqual(appointments[0].visit_code_sequence, 0)
        self.assertEqual(appointments[0].visit_schedule_name, "visit_schedule1")
        self.assertEqual(appointments[0].schedule_name, "schedule1")
        visit_schedule = site_visit_schedules.get_visit_schedule(
            appointments[0].visit_schedule_name
        )
        schedule = visit_schedule.schedules.get(appointments[0].schedule_name)
        self.assertEqual(schedule.visits.first.timepoint, Decimal("0.0"))
        self.assertEqual(appointments[0].timepoint, Decimal("0.0"))

    def test_baseline_appt_ontime_ok(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        form_validator = AppointmentFormValidator(
            cleaned_data=dict(
                subject_identifier=self.subject_identifier,
                appt_timing=ONTIME_APPT,
                appt_reason=SCHEDULED_APPT,
            ),
            instance=appointments[0],
        )
        try:
            form_validator.validate_appointment_timing()
        except ValidationError:
            self.fail("ValidationError unexpectedly raised")

    def test_baseline_appt_cannot_be_missed(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        form_validator = AppointmentFormValidator(
            cleaned_data=dict(
                subject_identifier=self.subject_identifier,
                appt_timing=MISSED_APPT,
                appt_reason=SCHEDULED_APPT,
            ),
            instance=appointments[0],
        )
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate_appointment_timing()
        self.assertIsNotNone(cm.exception)
        self.assertIn(INVALID_MISSED_APPT_NOT_ALLOWED_AT_BASELINE, form_validator._error_codes)

    def test_can_miss_scheduled_appt_if_not_baseline(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_timing, ONTIME_APPT)
        # create report for baseline visit
        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        appointments[0].refresh_from_db()

        self.assertEqual(appointments[1].visit_code_sequence, 0)
        self.assertEqual(appointments[1].timepoint, Decimal("1.0"))
        form_validator = AppointmentFormValidator(
            cleaned_data=dict(
                subject_identifier=self.subject_identifier,
                appt_timing=MISSED_APPT,
                appt_reason=SCHEDULED_APPT,
            ),
            instance=appointments[1],
        )

        try:
            form_validator.validate_appointment_timing()
        except ValidationError:
            self.fail("ValidationError unexpectedly raised")

    def test_cannot_miss_unscheduled_appt(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_timing, ONTIME_APPT)
        # create report for baseline visit
        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        appointments[0].refresh_from_db()

        # create unscheduled off of baseline appt
        appointment = self.create_unscheduled_appointment(appointments[0])
        self.assertEqual(appointment.visit_code, DAY01)
        self.assertEqual(appointment.visit_code_sequence, 1)
        self.assertEqual(appointment.timepoint, Decimal("0.0"))
        form_validator = AppointmentFormValidator(
            cleaned_data=dict(
                subject_identifier=self.subject_identifier,
                appt_timing=MISSED_APPT,
                appt_reason=UNSCHEDULED_APPT,
            ),
            instance=appointment,
        )

        with self.assertRaises(ValidationError) as cm:
            form_validator.validate_appointment_timing()
        self.assertIsNotNone(cm.exception)
        self.assertIn(INVALID_MISSED_APPT_NOT_ALLOWED, form_validator._error_codes)

    @override_settings(EDC_VISIT_TRACKING_ALLOW_MISSED_UNSCHEDULED=True)
    def test_can_miss_unscheduled_appt_if_allowed_in_settings(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_timing, ONTIME_APPT)
        # create report for baseline visit
        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        appointments[0].refresh_from_db()

        # create unscheduled off of baseline appt
        appointment = self.create_unscheduled_appointment(appointments[0])
        self.assertEqual(appointment.visit_code, DAY01)
        self.assertEqual(appointment.visit_code_sequence, 1)
        self.assertEqual(appointment.timepoint, Decimal("0.0"))
        form_validator = AppointmentFormValidator(
            cleaned_data=dict(
                subject_identifier=self.subject_identifier,
                appt_timing=MISSED_APPT,
                appt_reason=UNSCHEDULED_APPT,
            ),
            instance=appointment,
        )

        try:
            form_validator.validate_appointment_timing()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    @override_settings(
        SUBJECT_VISIT_MISSED_MODEL="edc_appointment_app.subjectvisitmissed",
        SUBJECT_VISIT_MISSED_REASONS_MODEL="edc_visit_tracking.subjectvisitmissedreasons",
    )
    def test_change_from_missed_removes_missed_visit_report(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_timing, ONTIME_APPT)
        # create report for baseline visit
        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        appointments[0].refresh_from_db()

        self.assertEqual(appointments[1].visit_code_sequence, 0)
        self.assertEqual(appointments[1].timepoint, Decimal("1.0"))
        appointments[1].appt_timing = MISSED_APPT
        appointments[1].appt_reason = SCHEDULED_APPT
        appointments[1].save_base(update_fields=["appt_timing", "appt_reason"])
        appointments[1].refresh_from_db()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[1],
            report_datetime=appointments[1].appt_datetime,
            reason=SCHEDULED,
        )

        # enter missed visit report
        get_subject_visit_missed_model_cls().objects.create(subject_visit=subject_visit)

        form_validator = AppointmentFormValidator(
            cleaned_data=dict(
                subject_identifier=self.subject_identifier,
                appt_timing=ONTIME_APPT,
                appt_reason=SCHEDULED_APPT,
            ),
            instance=appointments[1],
        )

        try:
            form_validator.validate_appointment_timing()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

        # assert missed visit report removed
        try:
            get_subject_visit_missed_model_cls().objects.get(subject_visit=subject_visit)
        except ObjectDoesNotExist:
            pass
        else:
            self.fail("ObjectDoesNotExist not raised")

    @override_settings(
        SUBJECT_VISIT_MISSED_MODEL="edc_appointment_app.subjectvisitmissed",
        SUBJECT_VISIT_MISSED_REASONS_MODEL="edc_viist_tracking.subjectvisitmissedreasons",
    )
    def test_change_to_missed_not_allowed_if_crfs_exist(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_timing, ONTIME_APPT)
        # create report for baseline visit
        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        appointments[0].refresh_from_db()

        self.assertEqual(appointments[1].visit_code_sequence, 0)
        self.assertEqual(appointments[1].timepoint, Decimal("1.0"))
        appointments[1].appt_timing = ONTIME_APPT
        appointments[1].appt_reason = SCHEDULED_APPT
        appointments[1].save_base(update_fields=["appt_timing", "appt_reason"])
        appointments[1].refresh_from_db()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[1],
            report_datetime=appointments[1].appt_datetime,
            reason=SCHEDULED,
        )

        # enter CRF
        crf_metadata = CrfMetadata.objects.create(
            model="edc_appointment.crfone",
            subject_identifier=subject_visit.subject_identifier,
            visit_schedule_name=subject_visit.visit_schedule_name,
            schedule_name=subject_visit.schedule_name,
            visit_code=subject_visit.visit_code,
            visit_code_sequence=subject_visit.visit_code_sequence,
            entry_status=KEYED,
            show_order=100,
        )

        form_validator = AppointmentFormValidator(
            cleaned_data=dict(
                subject_identifier=self.subject_identifier,
                appt_timing=MISSED_APPT,
                appt_reason=SCHEDULED_APPT,
            ),
            instance=appointments[1],
        )

        with self.assertRaises(ValidationError) as cm:
            form_validator.validate_appointment_timing()
        self.assertIsNotNone(cm.exception)
        self.assertIn(INVALID_APPT_TIMING_CRFS_EXIST, form_validator._error_codes)

        # reset CRF metadata
        crf_metadata.entry_status = REQUIRED
        crf_metadata.save()
        try:
            form_validator.validate_appointment_timing()
        except ValidationError:
            self.fail("ValidationError raised")

    @override_settings(
        SUBJECT_VISIT_MISSED_MODEL="edc_appointment_app.subjectvisitmissed",
        SUBJECT_VISIT_MISSED_REASONS_MODEL="edc_visit_tracking.subjectvisitmissedreasons",
    )
    def test_change_to_missed_not_allowed_if_requisitions_exist(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_timing, ONTIME_APPT)
        # create report for baseline visit
        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        appointments[0].refresh_from_db()

        self.assertEqual(appointments[1].visit_code_sequence, 0)
        self.assertEqual(appointments[1].timepoint, Decimal("1.0"))
        appointments[1].appt_timing = ONTIME_APPT
        appointments[1].appt_reason = SCHEDULED_APPT
        appointments[1].save_base(update_fields=["appt_timing", "appt_reason"])
        appointments[1].refresh_from_db()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[1],
            report_datetime=appointments[1].appt_datetime,
            reason=SCHEDULED,
        )

        # enter Requisition
        requisition_metadata = RequisitionMetadata.objects.create(
            model="edc_appointment.subjectrequisition",
            panel_name="panel1",
            subject_identifier=subject_visit.subject_identifier,
            visit_schedule_name=subject_visit.visit_schedule_name,
            schedule_name=subject_visit.schedule_name,
            visit_code=subject_visit.visit_code,
            visit_code_sequence=subject_visit.visit_code_sequence,
            entry_status=KEYED,
            show_order=200,
        )

        form_validator = AppointmentFormValidator(
            cleaned_data=dict(
                subject_identifier=self.subject_identifier,
                appt_timing=MISSED_APPT,
                appt_reason=SCHEDULED_APPT,
            ),
            instance=appointments[1],
        )

        with self.assertRaises(ValidationError) as cm:
            form_validator.validate_appointment_timing()
        self.assertIsNotNone(cm.exception)
        self.assertIn(INVALID_APPT_TIMING_REQUISITIONS_EXIST, form_validator._error_codes)

        # reset Requisition metadata
        requisition_metadata.entry_status = REQUIRED
        requisition_metadata.save()
        try:
            form_validator.validate_appointment_timing()
        except ValidationError:
            self.fail("ValidationError raised")

    def test_baseline_visit_report_datetime_must_match_appt_datetime(self):
        """Baseline appt date resets starting appointment, so visit
        report must match.
        """
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_timing, ONTIME_APPT)
        # create report for baseline visit
        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

from django.db import models, transaction
from django.db.models.deletion import ProtectedError
from edc_visit_schedule import site_visit_schedules

if TYPE_CHECKING:
    from .models import Appointment


class AppointmentDeleteError(Exception):
    pass


class AppointmentManagerError(Exception):
    pass


class SubjectOnScheduleError(Exception):
    pass


class AppointmentManager(models.Manager):

    use_in_migrations = True

    def get_by_natural_key(
        self,
        subject_identifier,
        visit_schedule_name,
        schedule_name,
        visit_code,
        visit_code_sequence,
    ) -> Appointment:
        return self.get(
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule_name,
            schedule_name=schedule_name,
            visit_code=visit_code,
            visit_code_sequence=visit_code_sequence,
        )

    @staticmethod
    def get_query_options(**kwargs) -> Dict[Any]:
        """Returns a dictionary or options.

        Dictionary is based on the appointment instance or everything
        else.
        """
        appointment = kwargs.get("appointment")
        schedule_name = kwargs.get("schedule_name")
        subject_identifier = kwargs.get("subject_identifier")
        visit_schedule_name = kwargs.get("visit_schedule_name")
        options: Dict[Any] = dict(visit_code_sequence=0)
        try:
            options.update(
                subject_identifier=appointment.subject_identifier,
                visit_schedule_name=appointment.visit_schedule_name,
                schedule_name=appointment.schedule_name,
            )
        except AttributeError:
            options.update(subject_identifier=subject_identifier)
            try:
                visit_schedule_name, schedule_name = visit_schedule_name.split(".")
                options.update(
                    visit_schedule_name=visit_schedule_name, schedule_name=schedule_name
                )
            except ValueError:
                options.update(visit_schedule_name=visit_schedule_name)
            except AttributeError:
                pass
            if schedule_name and not visit_schedule_name:
                raise TypeError(
                    f"Expected visit_schedule_name for schedule_name "
                    f"'{schedule_name}'. Got {visit_schedule_name}"
                )
            elif schedule_name:
                options.update(schedule_name=schedule_name)
        return options

    @staticmethod
    def get_visit_code(action, schedule, **kwargs) -> str | None:
        """Updates the options dictionary with the next or previous
        visit code in the schedule.

        if both visit_code and appointment are in kwargs visit_code
        takes precedence over apppointment.visit_code
        """
        visit_code = kwargs.get("visit_code")
        if not visit_code:
            try:
                appointment = kwargs.get("appointment")
                visit_code = appointment.visit_code
            except AttributeError:
                pass
        if action == "next":
            visit = schedule.visits.next(visit_code)
        elif action == "previous":
            visit = schedule.visits.previous(visit_code)
        else:
            raise AppointmentManagerError(
                f"Unknown action. Expected one of [next, previous]. Got '{action}'."
            )
        try:
            visit_code = visit.code
        except AttributeError:
            visit_code = None
        return visit_code

    def first_appointment(self, **kwargs) -> Appointment | None:
        """Returns the first appointment instance for the given criteria.

        For visit_code_sequence=0.

        For example:
            first_appointment = (Appointment.objects.
                first_appointment(appointment=appointment))
        or:
            first_appointment = Appointment.objects.first_appointment(
                subject_identifier=subject_identifier,
                visit_schedule_name=visit_schedule_name,
                schedule_name=schedule_name)
        """
        options = self.get_query_options(**kwargs)
        try:
            first_appointment = self.filter(**options).order_by(
                "timepoint", "visit_code_sequence"
            )[0]
        except IndexError:
            first_appointment = None
        return first_appointment

    def last_appointment(self, **kwargs) -> Appointment | None:
        """Returns the last appointment relative to the criteria.

        For visit_code_sequence=0.
        """
        options = self.get_query_options(**kwargs)
        try:
            last_appointment = [
                obj
                for obj in self.filter(**options).order_by("timepoint", "visit_code_sequence")
            ][-1]
        except IndexError:
            last_appointment = None
        return last_appointment

    def next_appointment(self, **kwargs) -> Appointment | None:
        """Returns the next appointment relative to the criteria or
        None if there is no next.

        For visit_code_sequence=0.

        Next is the next visit in a schedule, so schedule_name is required.

        For example:
            next_appointment = (Appointment.objects.
                first_appointment(appointment=appointment))
        or:
            next_appointment = Appointment.objects.first_appointment(
                visit_code=visit_code, appointment=appointment)
        or:
            next_appointment = Appointment.objects.first_appointment(
                visit_code=visit_code,
                subject_identifier=subject_identifier,
                visit_schedule_name=visit_schedule_name,
                schedule_name=schedule_name)
        """
        options = self.get_query_options(**kwargs)
        schedule = site_visit_schedules.get_visit_schedule(
            options.get("visit_schedule_name")
        ).schedules.get(options.get("schedule_name"))
        options.update(visit_code=self.get_visit_code("next", schedule, **kwargs))
        try:
            next_appointment = self.filter(**options).order_by(
                "timepoint", "visit_code_sequence"
            )[0]
        except IndexError:
            next_appointment = None
        return next_appointment

    def previous_appointment(self, **kwargs) -> Appointment | None:
        """Returns the previous appointment relative to the criteria
        or None if there is no previous.

        For visit_code_sequence=0.
        """
        options = self.get_query_options(**kwargs)
        schedule = site_visit_schedules.get_visit_schedule(
            options.get("visit_schedule_name")
        ).schedules.get(options.get("schedule_name"))
        options.update(visit_code=self.get_visit_code("previous", schedule, **kwargs))
        try:
            previous_appointment = (
                self.filter(**options)
                .order_by("timepoint", "visit_code_sequence")
                .reverse()[0]
            )
        except IndexError:
            previous_appointment = None
        return previous_appointment

    def delete_for_subject_after_date(
        self,
        subject_identifier=None,
        cutoff_datetime=None,
        op=None,
        visit_schedule_name=None,
        schedule_name=None,
        is_offstudy=None,
    ) -> int:
        """Deletes appointments for a given subject_identifier with
        appt_datetime greater than `dt`.

        If a visit form exists for any appointment, a ProtectedError will
        be raised.
        """
        # validate "op"
        valid_ops = ["gt", "gte"]
        if op and op not in valid_ops:
            formatted = ", ".join(valid_ops)
            raise TypeError(f"Allowed lookup operators are {formatted}. Got {op}.")
        op = "gte" if op is None else op

        # prepare options
        options = {
            "subject_identifier": subject_identifier,
            f"appt_datetime__{op}": cutoff_datetime,
        }
        if not is_offstudy:
            try:
                visit_schedule_name, schedule_name = visit_schedule_name.split(".")
            except (ValueError, AttributeError):
                pass
            if not schedule_name or not visit_schedule_name:
                raise AppointmentManagerError(
                    f"Expected both the visit_schedule_name and schedule_name. "
                    f"Got schedule_name='{schedule_name}', "
                    f"visit_schedule_name='{visit_schedule_name}'"
                )
            options.update(dict(visit_schedule_name=visit_schedule_name))
            options.update(dict(schedule_name=schedule_name))

        # delete future appointments until the first with a
        # visit report
        deleted = 0
        appointments = self.filter(**options).order_by("timepoint", "visit_code_sequence")
        for appointment in appointments.reverse():
            try:
                with transaction.atomic():
                    appointment.delete()
                    deleted += 1
            except ProtectedError:
                break
            except AppointmentDeleteError:
                pass
        return deleted



from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist

from ..constants import (
    CANCELLED_APPT,
    COMPLETE_APPT,
    IN_PROGRESS_APPT,
    INCOMPLETE_APPT,
    NEW_APPT,
)
from ..models import Appointment
from ..utils import update_unscheduled_appointment_sequence


class AppointmentViewMixin:

    """A view mixin to handle appointments on the dashboard."""

    appointment_model_wrapper_cls = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._appointment = None
        self._appointments = None
        self._wrapped_appointments = None
        self.appointment_model = "edc_appointment.appointment"
        self.appointment_id = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        self.appointment_id = kwargs.get("appointment")
        update_unscheduled_appointment_sequence(
            subject_identifier=self.kwargs.get("subject_identifier"),
        )
        has_call_manager = True if django_apps.app_configs.get("edc_call_manager") else False
        context.update(
            appointment=self.appointment_wrapped,
            appointments=self.appointments_wrapped,
            CANCELLED_APPT=CANCELLED_APPT,
            COMPLETE_APPT=COMPLETE_APPT,
            INCOMPLETE_APPT=INCOMPLETE_APPT,
            IN_PROGRESS_APPT=IN_PROGRESS_APPT,
            NEW_APPT=NEW_APPT,
            has_call_manager=has_call_manager,
        )
        return context

    @property
    def appointment_options(self):
        opts = {}
        if self.kwargs.get("appointment"):
            opts = dict(id=self.kwargs.get("appointment"))
        elif (
            self.kwargs.get("subject_identifier")
            and self.kwargs.get("visit_schedule_name")
            and self.kwargs.get("schedule_name")
            and self.kwargs.get("visit_code")
        ):
            visit_code_sequence = self.kwargs.get("visit_code_sequence") or 0
            opts = dict(
                subject_identifier=self.kwargs.get("subject_identifier"),
                visit_schedule_name=self.kwargs.get("visit_schedule_name"),
                schedule_name=self.kwargs.get("schedule_name"),
                visit_code=self.kwargs.get("visit_code"),
                visit_code_sequence=visit_code_sequence,
            )
        return opts

    @property
    def appointment(self):
        if not self._appointment:
            if self.appointment_id:
                try:
                    self._appointment = Appointment.objects.get(id=self.appointment_id)
                except ObjectDoesNotExist:
                    opts = self.appointment_options
                    if opts:
                        try:
                            self._appointment = self.appointment_model_cls.objects.get(**opts)
                        except ObjectDoesNotExist:
                            pass
        return self._appointment

    @property
    def appointment_wrapped(self):
        if self.appointment:
            return self.appointment_model_wrapper_cls(model_obj=self.appointment)
        return None

    @property
    def appointments(self):
        """Returns a Queryset of all appointments for this subject."""
        if not self._appointments:
            self._appointments = self.appointment_model_cls.objects.filter(
                subject_identifier=self.subject_identifier
            ).order_by("timepoint", "visit_code_sequence")
        return self._appointments

    @property
    def appointments_wrapped(self):
        """Returns a list of wrapped appointments."""
        if not self._wrapped_appointments:
            if self.appointments:
                wrapped = [
                    self.appointment_model_wrapper_cls(model_obj=obj)
                    for obj in self.appointments
                ]
                for i in range(0, len(wrapped)):
                    if wrapped[i].appt_status == IN_PROGRESS_APPT:
                        wrapped[i].disabled = False
                        for j in range(0, len(wrapped)):
                            if j != i:
                                wrapped[j].disabled = True
                self._wrapped_appointments = wrapped
        return self._wrapped_appointments or []

    @property
    def appointment_model_cls(self):
        return django_apps.get_model(self.appointment_model)

    def empty_appointment(self, **kwargs):
        return self.appointment_model_cls()
